import lasio as laua
import io
import numpy as np
import pandas as pd

def datasets_to_las(path, datasets, custom_units=None, **kwargs):
    """
    copyright 2021 Agile Scientific
    license
    Apache 2.0
    """
    """
    Write datasets to a LAS file on disk.

    Args:
        path (Str): Path to write LAS file to. If None, returns string buffer.
        datasets (Dict['<name>': pd.DataFrame]): Dictionary maps a
            dataset name (e.g. 'Curves') or 'Header' to a pd.DataFrame.
        custom_units (Dict[str, str], optional): Dictionary mapping curve names to their units.
            If a curve's unit is not specified, it defaults to an empty string.
    Returns:
        str if path is None, otherwise writes to file
    """
    from functools import reduce
    import warnings
    from lasio import HeaderItem, CurveItem, SectionItems
    if custom_units is None:
        custom_units={}
    # ensure no NaN values in header
    if 'Header' in datasets:
        datasets['Header'] = datasets['Header'].fillna('')
    
    # instantiate new LASFile to parse data & header to
    las = laua.LASFile()

    # set header df as variable to later retrieve curve meta data from
    header = datasets['Header']
    
    extracted_units = {}
    if not header.empty:
        curve_header = header[header['section'] == 'Curves']
        for _, row in curve_header.iterrows():
            if row['unit'] and not pd.isna(row['unit']):  # Check for NaN
                extracted_units[str(row['original_mnemonic'])] = str(row['unit'])

    # Combine extracted units with custom units, custom units take precedence
    all_units = {**extracted_units, **custom_units}
    
    column_fmt = {}
    for curve in las.curves:
        column_fmt[curve.mnemonic] = "%10.5f"
    
    # unpack datasets
    for dataset_name, df in datasets.items():
        # dataset is the header
        if dataset_name == 'Header':
            # parse header pd.DataFrame to LASFile
            for section_name in set(df.section.values):
                # get header section df
                df_section = df[df.section == section_name]

                if section_name == 'Curves':
                    pass  # curves header items are handled in curve data loop
                
                elif section_name == 'Version':
                    for mnem in ['VERS', 'WRAP', 'DLM']:
                        section_data = df_section[df_section.original_mnemonic == mnem]
                        if len(section_data) > 0:
                            setattr(las.version, mnem, str(section_data['value'].values[0]))

                elif section_name in ['Well', 'Parameter']:
                    las.sections[section_name] = SectionItems(
                        [HeaderItem(str(r.original_mnemonic),
                                  str(r.unit) if not pd.isna(r.unit) else '',
                                  str(r.value) if not pd.isna(r.value) else '',
                                  str(r.descr) if not pd.isna(r.descr) else '')
                         for _, r in df_section.iterrows()
                         if not pd.isna(r.original_mnemonic)])  # Skip NaN mnemonics

                elif section_name == 'Other':
                    las.sections["Other"] = str(df_section['descr'].iloc[0]) if not pd.isna(df_section['descr'].iloc[0]) else ''

                else:
                    warnings.warn(f"LAS Section was not recognized: '{section_name}'", stacklevel=2)

        # dataset contains curve data
        if dataset_name in ['Curves', 'Formation', 'Core']:  # explicit curve sections
            for column_name in df.columns:
                curve_data = df[column_name]
                curve_unit = all_units.get(str(column_name), '')
                las.append_curve(mnemonic=str(column_name),
                               data=curve_data,
                               unit=str(curve_unit),
                               descr='',
                               value='')

    # Handle NULL value
    try:
        null_value = float(header[header.original_mnemonic == 'NULL'].value.iloc[0])
    except (IndexError, ValueError):
        null_value = -999.25
    
    # Set the NULL value in both the well section and as general null_value
    las.well['NULL'] = HeaderItem('NULL', '', null_value, 'Null value')
    las.null_value = null_value

    # Set column formatter if not provided
    if 'column_fmt' not in kwargs:
        kwargs['column_fmt'] = column_fmt
        
    # write file to disk or return string
    if path is not None:
        with open(path, mode='w') as f:
            las.write(f, **kwargs)

    buffer = io.StringIO()
    las.write(buffer, **kwargs)
    buffer.seek(0)
    return buffer.getvalue()



"""
Copyright (c) 2024-2025 ROCK LAB PRIVATE LIMITED
This file is part of "Stresslog" project and is released under the
GNU Affero General Public License v3.0 (AGPL-3.0)
See the GNU Affero General Public License for more details: <https://www.gnu.org/licenses/agpl-3.0.html>
"""
import pytest
import stresslog as lst
import pandas as pd
import numpy as np

def test_integration_basic():
# Test case 1: Basic well log generation and plotting
    attrib = [10, 0, 0, 0, 0, 0, 0, 0] #KB, GL, 
    well = lst.create_random_well(kb=35, gl=-200, step=5)
    print(well.df())
    dev = lst.getwelldev(wella=well)
    print(dev.df())
    output = lst.compute_geomech(dev, writeFile=False)

    assert output is not None, "plotPPzhang did not return an output"

def test_integration_doi():
    # Test case 2: Well with attributes and DOI
    attrib = [10, -200, 0, 0, 0, 0, 0, 0]
    well = lst.create_random_well(kb=312, gl=300, step=4)
    dev = lst.getwelldev(wella=well, kickoffpoint=3000, final_angle=90, rateofbuild=0.2, azimuth=270)
    output = lst.compute_geomech(dev, attrib=attrib, doi=2625, writeFile=False)
    assert output is not None, "plotPPzhang with attributes and DOI did not return an output"

def test_integration_advanced():
    # Test case 3: Shallow angle well and custom TECB. now with Formations, Imagelog and UCS data

    ucs = pd.read_csv('https://raw.githubusercontent.com/GeoArkadeep/supporting-data-for-EOS-Northern-Lights/main/UCSdata.csv')
    formations = pd.read_csv('https://raw.githubusercontent.com/GeoArkadeep/supporting-data-for-EOS-Northern-Lights/main/NorthernLights-31_5-7.csv')
    imagelog = pd.read_csv('https://raw.githubusercontent.com/GeoArkadeep/supporting-data-for-EOS-Northern-Lights/main/31_5-7_Image.csv')
    
    test_m_depth=2000
    attrib = [10, -200, 0, 0, 0, 0, 0, 0]
    well = lst.create_random_well(kb=35, gl=-200, step=4, seed=33)
    dev = lst.getwelldev(wella=well, kickoffpoint=300, final_angle=10, rateofbuild=0.2, azimuth=270)
    output = lst.compute_geomech(dev, attrib=attrib, doi=2625, tecb=0.55, display=True, forms = formations, UCSs = ucs, flags = imagelog, debug = True)
    
    index = lst.find_nearest_depth(output[0]['MD'].to_numpy(),test_m_depth)[0]
    
    assert abs(output[0]['OVERBURDEN_PRESSURE'].to_numpy()[index] - 5309.732562366982) < 0.001, f"Overburden check fail: {output[0]['OVERBURDEN_PRESSURE'].to_numpy()[index]}"
    print(f"Overburden check passed: {output[0]['OVERBURDEN_PRESSURE'].to_numpy()[index]}psi")
    
    assert abs(output[0]['FracPressure'].to_numpy()[index] - 9402.524110909559) < 0.001, f"Frac Pressure check fail: {output[0]['FracPressure'].to_numpy()[index]}"
    print(f"Frac Pressure check passed: {output[0]['FracPressure'].to_numpy()[index]}psi")

def test_analog():
    #Test case 4: Test analog well
    formations = pd.read_csv('https://raw.githubusercontent.com/GeoArkadeep/supporting-data-for-EOS-Northern-Lights/main/NorthernLights-31_5-7.csv')
    well = lst.create_random_well(kb=35, gl=-200, step=4, seed=33)
    dev = lst.getwelldev(wella=well, kickoffpoint=300, final_angle=10, rateofbuild=0.2, azimuth=270)
    
    new_forms = formations.rename(columns={"Top TVD": "TopTVD", "Formation Name": "Name"})[["TopTVD", "Name"]]
    target_forms = formations.rename(columns={"Top TVD": "TopTVD", "Formation Name": "Name"}).assign(TopTVD=lambda df: df["TopTVD"] + 150)[["TopTVD", "Name"]]
    analogwell,analoglas, analogdev = lst.get_analog(dev,new_forms, target_forms, kb=40,gl=-200,dev=None,kop=0,ma=0,rob=0.1,azi=0,debug=False,td=5000)
    
    df_analog = analogwell.df()
    dev_df = dev.df() # 'dev' is the source well passed to lst.get_analog

    idx_orig = lst.find_nearest_depth(dev_df['TVDM'].to_numpy(), 2900.0)[0]
    gr_orig = dev_df['GR'].iloc[idx_orig]
    rhob_orig = dev_df['RHOB'].iloc[idx_orig]

    idx_analog = lst.find_nearest_depth(df_analog['TVDM'].to_numpy(), 3050.0)[0]
    gr_analog = df_analog['GR'].iloc[idx_analog]
    rhob_analog = df_analog['RHOB'].iloc[idx_analog]

    # Seed(33) makes random logs deterministic. Allow small tolerance for interpolations.
    assert np.isclose(gr_orig, gr_analog, rtol=1e-2),f"GR mismatch after depth shift: original GR={gr_orig} at TVDM~2900, analog GR={gr_analog} at TVDM~3050"
    print(f"GR match after depth shift: original GR={gr_orig} at TVDM~2900, analog GR={gr_analog} at TVDM~3050")
    assert np.isclose(rhob_orig, rhob_analog, rtol=1e-2),f"RHOB mismatch after depth shift: original RHOB={rhob_orig} at TVDM~2900, analog RHOB={rhob_analog} at TVDM~3050"
    print(f"RHOB match after depth shift: original RHOB={rhob_orig} at TVDM~2900, analog RHOB={rhob_analog} at TVDM~3050")
    
    


import io
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
from difflib import SequenceMatcher

import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.simplefilter(action='ignore', category=RuntimeWarning)

from .syntheticLogs import getwelldev, create_random_well, create_header
from .geomechanics import remove_curves, add_curves, find_TVD, compute_geomech
from .thirdparty import datasets_to_las


#from stresslog import *


def create_blank_well(kb,gl,kop=0, maxangle=0, rob=0.1,azimuth=0,dev=None, spacing=0.15,td=5000):
    """Create a random well object with specified parameters.

    Parameters
    ----------
    kb : float
        Kelly bushing elevation in meters
    gl : float
        Ground level elevation in meters
    kop : float, optional
        Kickoff point depth in meters, by default 0
    maxangle : float, optional
        Maximum deviation angle in degrees, by default 0
    step : float, optional
        Depth step size in meters, by default 0.15
    drop : list, optional
        List of curves to exclude from output, by default []

    Returns
    -------
    welly.Well
        Well object containing randomly generated log data
    """
    blankwell = getwelldev(wella=create_random_well(kb=kb, gl=gl, stopper=td, step=spacing),step=spacing, deva=dev, kickoffpoint=kop, final_angle=maxangle, rateofbuild=rob, azimuth=azimuth)
    blankwell = remove_curves(blankwell,["DTCO","RHOB","NPHI","GR","ILD","DTS","NU"])
    return blankwell
    
    


def uniform_resample(df, key, step):
    # Save the original index name
    originalindexkey = df.index.name
    if originalindexkey:
        df = df.reset_index()
    df = df.set_index(key)
    # Create new uniform index
    uniform_spacing = step
    new_index = np.arange(df.index.min(), df.index.max() + uniform_spacing, uniform_spacing)

    # Resample using interpolation
    df = df.reindex(df.index.union(new_index)).interpolate(method="index").reindex(new_index)
    df = df.reset_index()
    if originalindexkey:
        df = df.set_index(originalindexkey)

    return df
    

def depth_shift_process(df, current_forms_dict, new_forms_dict):
    """
    Transform a dataframe by remapping its index according to provided mapping dictionaries.
    
    Parameters:
    -----------
    df : pandas.DataFrame
        Input dataframe with numerical index
    current_forms_dict : dict
        Dictionary mapping form names to current index positions
    new_forms_dict : dict
        Dictionary mapping form names to desired new index positions
    
    Returns:
    --------
    pandas.DataFrame
        Transformed dataframe with new index positions
    """
    #df = df.fillna(0.0)
    offset = (new_forms_dict[list(new_forms_dict)[-1]]-current_forms_dict[list(current_forms_dict)[-1]])
    print(offset)
    current_forms_dict["Fin"] = float(df.index[-1])
    new_forms_dict["Fin"] = float(df.index[-1]+offset)
    print(current_forms_dict)
    print(new_forms_dict)
    
    # Convert dictionaries to sorted lists of tuples for easier processing
    current_points = sorted(current_forms_dict.items(), key=lambda x: x[1])
    new_points = sorted(new_forms_dict.items(), key=lambda x: x[1])
    
    # Extract x and y coordinates for the mapping
    current_x = [0] + [p[1] for p in current_points]  # Add 0 as starting point
    new_x = [0] + [p[1] for p in new_points]  # Add 0 as starting point

    # Create interpolation function
    transform_func = interp1d(current_x, new_x, kind='linear', bounds_error=False, 
                            fill_value=(new_x[0], new_x[-1]))
    
    # Transform the index
    new_index = transform_func(df.index)
    
    # Create new dataframe with transformed index
    df_new = df.copy()
    df_new.index = new_index
    
    # Sort by new index to ensure proper ordering
    df_new = df_new.sort_index()
    
    # Create regular spacing between min and max of new index
    num_points = len(df_new)
    regular_index = pd.Index(np.linspace(df_new.index.min(), df_new.index.max(), num_points))
    
    # Create a new dataframe with the regular index
    df_regular = pd.DataFrame(index=regular_index, columns=df_new.columns)
    
    # Copy data to the new dataframe and interpolate
    for column in df_new.columns:
        df_regular[column] = np.interp(
            regular_index,
            df_new.index,
            df_new[column].values,
            left=df_new[column].iloc[0],
            right=df_new[column].iloc[-1]
        )
    df_regular.index.names = df.index.names
    return df_regular


def get_analog_well(well,cfd,nfd,spacing=1,blankwell=None, indices="TVDM", dev= None, kop= 0, rob= 0, maxangle= 0, azimuth=0, name= "Analog", kb=0, gl=0, debug=False):
    #returns welly.well object with valid deviation data
    # Define mapping dictionaries
    df = well.df()
    df["DEPT"] = df.index.values
    if debug:
        print(df)
    if indices is not None:
        df = df.set_index(indices)
    if debug:
        print(df)
    #Resample to regular spacing, eg 1m
    uniform_spacing = spacing

    # Create new uniform index
    new_index = np.arange(df.index.min(), df.index.max() + uniform_spacing, uniform_spacing)

    # Resample using interpolation
    # Method can be 'linear', 'cubic', 'nearest', etc.
    df = df.reindex(df.index.union(new_index)).interpolate(method='index').reindex(new_index)

    #df_resampled = df.interpolate(method="index")

    if debug:
        print(df)
    

    # Transform the dataframe
    dft = depth_shift_process(df, cfd, nfd)
    dft=dft.drop(["MD"],axis=1) #MD and TVD (as well as all well specific data is invalid now 
    dft=dft.drop(["DEPT"],axis=1) #so we remove these to prevent confusion
    dftvd = uniform_resample(dft,"TVDM",spacing)
    if debug:
        print("dft fresh")
        print(dftvd)
    
    #Now we create the backbone of the proposed well with the new proposed deviation data
    if blankwell is None:
        if dev is not None:
            blankwell = getwelldev(wella=create_random_well(kb=kb, gl=gl, step=spacing),step=spacing,deva=dev)
        else:
            blankwell = getwelldev(wella=create_random_well(kb=kb, gl=gl, step=spacing),step=spacing, kickoffpoint=kop, final_angle=maxangle, rateofbuild=rob, azimuth=azimuth)
        blankwell = remove_curves(blankwell,["DTCO","RHOB","NPHI","GR","ILD","DTS","NU"])
    blankdf = blankwell.df()
    analogdevnp = blankwell.location.deviation
    analogdevdf = pd.DataFrame(analogdevnp, columns=["MD", "INC", "AZIM"])

    print(analogdevdf)
    analogdevdf.to_csv("analogdevdf.csv",index=False)
    kb = float(blankwell.header.set_index("mnemonic").loc["KB", "value"])
    gl = float(blankwell.header.set_index("mnemonic").loc["GL", "value"])
    #blankdf = uniform_resample(blankdf,"TVDM",spacing)
    if debug:
        print("Blankdf:")
        print(blankdf)
    dfmd=blankdf
    # Ensure dftvd's index is float and sorted for interpolation
    dftvd.index = dftvd.index.astype(float)
    dftvd = dftvd.sort_index()

    xp = dftvd.index.to_numpy()  # Independent variable from dftvd (TVD)
    fp = dftvd.to_numpy()  # Dependent variable(s) from dftvd (values to interpolate)

    x = dfmd["TVDM"].to_numpy()  # Values to interpolate at (TVDM from dfmd)

    # Perform interpolation using numpy.interp
    # np.interp handles 1D arrays, so we do this for each column in dftvd
    interpolated_values = np.vstack([
        np.interp(x, xp, fp[:, col_idx]) for col_idx in range(fp.shape[1])
    ]).T

    # Create a new dataframe with interpolated rows
    dft_new = pd.DataFrame(interpolated_values, index=dfmd.index, columns=dftvd.columns)

    # Optionally, include MD from dfmd for context
    dft_new["MD"] = dfmd["MD"]
    dft_new["TVDM"] = dfmd["TVDM"]
    dft_new["DEPT"] = dfmd["MD"]
    #dft_new.index.name = "DEPT"

    if debug:
        print("dft_new:")
        print(dft_new)
    
    analogwell = add_curves(blankwell,dft_new,True)
    if debug:
        print("analogwell's df")
        print(analogwell.df())

        #returning the analogwell directly is a bad idea, let's write it to a las string and return that instead
        #this way we can i) return something industry standard, ii) human readable so we can check the outcome
    
        print("Kellybushing of target well is: ",kb)
    hdf = create_header(
        name="analogwell",#f"{well_name}-{key}", 
        uwi="ANALOGWELL",#f"WELL#{key}", 
        strt=str(dft_new["MD"].iloc[0]),#str(starter), 
        stop=str(dft_new["MD"].iloc[-1]),#str(stopper), 
        step=str(dft_new["MD"].iloc[-5]-dft_new["MD"].iloc[-4]),#str(stepper),
        kb=str(kb),
        gl=str(gl),
    )
    if debug:
        print(hdf)
    #analogwell.header = hdf
    cols = list(dft_new.columns)
    cols.reverse()

    dft_new = dft_new[cols]
    if debug:
        print(dft_new)
    las_string = datasets_to_las("test.las" if debug else None, {'Curves': dft_new, 'Header': hdf}, 
                             {})
    #print(las_string)
    if debug:
        print(dft_new)
        print(analogwell)
        print(analogwell.df()["MD"])
        
        
        dft.reset_index(inplace=True)
        df.reset_index(inplace=True)
        dft.set_index("TVDM",inplace=True)
        df.set_index("TVDM",inplace=True)
        
        # Create the plot
        import matplotlib.pyplot as plt
        plt.plot(dftvd['DTHM'],dftvd.index.values,label="TVD Model", alpha=0.7)
        plt.plot(df['DTHM'],df.index.values,label="Post-drill well", alpha=0.7)
        plt.plot(dft_new['DTHM'],dft_new["TVDM"],label="Analog well")
        plt.legend()
        # Force plot to start at 0,0
        plt.xlim(240, 0)
        plt.ylim(plt.ylim()[1],0)
        # Add ticks on all edges
        ax = plt.gca()
        ax.tick_params(which='both', direction='out')  # Make ticks point inward
        ax.yaxis.set_ticks_position('both')  # Show ticks on both left and right
        ax.xaxis.set_ticks_position('both')  # Show ticks on both top and bottom
        
        # Set major and minor grid
        major_ticks = np.arange(0, max(max(dft.index.values), max(dft.index.values)) + 1000, 1000)
        minor_ticks = np.arange(0, max(max(dft.index.values), max(dft.index.values)) + 100, 100)


        ax.set_yticks(major_ticks)
        ax.set_yticks(minor_ticks, minor=True)
        # Add grid
        plt.grid(which='major', color='gray', linestyle='-', alpha=0.5)
        plt.grid(which='minor', color='gray', linestyle=':', alpha=0.3)

        # Save the figure
        plt.savefig("depthshiftprocessing.png")
        plt.close()
    analogwell.header = analogwell.header.drop(index=0).reset_index(drop=True)
    return analogwell,io.StringIO(las_string), analogdevdf



def fuzzymatch(df, target_key):
    """
    Find the closest matching column name from a DataFrame, regardless of similarity.

    Args:
        df (pd.DataFrame): The DataFrame containing the columns.
        target_key (str): The target column name to match.

    Returns:
        str: The closest matching column name from the DataFrame.
    """
    def similarity(a, b):
        return SequenceMatcher(None, a, b).ratio()
    
    # Find the column with the highest similarity score
    target_key_lower = target_key.lower()
    best_match = max(df.columns, key=lambda col: similarity(target_key_lower, col.lower()))

    return best_match


def convert_df_tvd(df,well):
    """
    Converts the first column of formation or other dataframes from MD to TVD, given a welly.Well object with valid deviation data.
    
    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe who's first column has the measured depths we want to convert.
    well : welly.Well
        Source well object containing the original log data and wellbore information.
        Must valid deviation data.
    Returns
    -------
    pandas.DataFrame
        The dataframe with the first column converted to TVD and column name set to TopTVD
    """
    
    originalkey = list(df)[0]
    df[originalkey] = df[originalkey].apply(lambda x: find_TVD(well, x))
    print(df)
    if originalkey != 'TopTVD':
        df = df.rename(columns={originalkey: 'TopTVD'})
    return df

def extract_depth_dict(df):
    originalkey = list(df)[0]
    f_names = ("Name")#fuzzymatch(df,"Name")
    return {row[f_names]: round(float(row[originalkey]), 2) for _, row in df.iterrows()}

def get_analog(well,current_forms, target_forms, kb,gl,dev=None,kop=0,ma=0,rob=0.1,azi=0,debug=False,td=5000):
    """
    Generate an analog well by transforming log data from an existing well according to specified formation depths
    and wellbore geometry parameters. This function creates a new well object with transformed log data that 
    maintains the character of the original logs while honoring new formation tops and a new wellbore trajectory.

    The function performs several key operations:
    1. Converts MD (Measured Depth) formations to TVD (True Vertical Depth) if necessary
    2. Creates a blank wellbore with specified geometry
    3. Transforms the original well's log data to match new formation depths
    4. Generates a new well object with the transformed data

    Parameters
    ----------
    well : welly.Well
        Source well object containing the original log data and wellbore information.
        Must contain basic log curves (e.g., GR, RHOB, etc.) and valid deviation data.

    current_forms : pandas.DataFrame
        DataFrame containing the current formation tops information.
        Must include columns for depth values and formation names.
        Depths can be in either MD or TVD (specified by column name).
        Example format:
            TopTVD/MD  Name
            1690      Alpha
            2380      Beta
            ...       ...

    target_forms : pandas.DataFrame
        DataFrame containing the target formation tops information for the analog well.
        Must follow the same format as current_forms.
        These depths represent where formations should appear in the analog well.

    kb : float
        Kelly Bushing elevation in meters for the analog well.
        Used as the reference point for depth measurements.

    gl : float
        Ground Level elevation in meters for the analog well.
        Must be less than or equal to kb if onshore and equal to negetive water depth if offshore

    dev : array-like, optional
        Custom deviation survey data for the analog well.
        If provided, overrides the kop, ma, rob, and azi parameters.
        Format should be compatible with welly.Well deviation data.
        Default is None.

    kop : float, optional
        Kick-Off Point depth in meters.
        Depth at which the wellbore begins to deviate from vertical.
        Must be >= 0.
        Default is 0.

    ma : float, optional
        Maximum angle (deviation) in degrees.
        The maximum inclination angle the wellbore will reach.
        Must be between 0 and 90.
        Default is 0 (vertical well).

    rob : float, optional
        Rate Of Build in degrees per meter.
        Rate at which the wellbore builds angle from the KOP.
        Typical values range from 0.1 to 3.0 degrees/meter.
        Default is 0.1.

    azi : float, optional
        Azimuth in degrees.
        The compass direction of the wellbore deviation.
        Must be between 0 and 360.
        Default is 0 (North).

    debug : bool, optional
        If True, enables debug mode which:
        - Prints intermediate dataframes
        - Generates diagnostic plots
        - Saves additional output files
        - Provides verbose console output
        Default is False.

    Returns
    -------
    tuple
        A tuple containing two elements:
        - welly.Well : The created analog well object with transformed log data
          and new trajectory information
        - io.StringIO : A StringIO object containing the LAS file representation
          of the analog well

    Notes
    -----
    - The function uses fuzzy matching to find the formation names column in the dataframes,
      so the column name can be "Name", "Formation_Name", etc.
    - Log data is resampled to maintain consistent spacing throughout the transformation
    - The function preserves the character and relationships between different log curves
      while shifting them to new depths
    - The transformation process interpolates depths linearly between formation tops, 
      while preserving thickness below the deepest formation top
    - Formation top names must be consistent between current_forms and target_forms
    - If a formation has pinched out, it should still be included in both dataframes
      with its depth matching either the overlying or underlying formation based on
      geological interpretation

    """
    offset1 = kb-gl
    offset0 = (float(well.header.set_index("mnemonic").loc["KB", "value"]))-float(well.header.set_index("mnemonic").loc["GL", "value"])
    
    step = well.df().index.values[1] - well.df().index.values[0]
    if debug:
        print(step)
    if list(current_forms)[0].lower()=="toptvd":
        cfd = extract_depth_dict(current_forms)
    else:
        cfd = extract_depth_dict(convert_df_tvd(current_forms,well))
        
    
    newblankwell=create_blank_well(kb=kb,gl=gl,spacing=step,maxangle=ma,kop=kop,azimuth=azi,dev=dev,td=td)
    
    if list(target_forms)[0].lower()=="toptvd":
        nfd = extract_depth_dict(target_forms)
    else:
        nfd = extract_depth_dict(convert_df_tvd(target_forms,newblankwell))

    cfd["Start"] = offset0
    nfd["Start"] = offset1
    
    return get_analog_well(well,cfd,nfd,5,blankwell=newblankwell,debug=debug)
    


if __name__=="__main__":
    import welly
    spacing = 5
    wellu = welly.Well.from_las("testlas.las")
    print(wellu)
    dev = pd.read_csv("testdev.csv")
    print(dev)

    well = getwelldev(wella=wellu,deva=dev,step=spacing)
    print(well.df())
    
    well.header = pd.concat([well.header[well.header["mnemonic"] != "KB"], pd.DataFrame({"mnemonic": ["KB"], "value": [80]})], ignore_index=True)
    well.header = pd.concat([well.header[well.header["mnemonic"] != "GL"], pd.DataFrame({"mnemonic": ["GL"], "value": [20]})], ignore_index=True)
    
    #well = getwelldev(wella=create_random_well(kb=35, gl=-200, step=spacing),step=spacing, kickoffpoint=2500, final_angle=25, rateofbuild=0.1, azimuth=270)
    form_md = """TopTVD,StratiNum,Name,GR_Cut,Struc_Top,Struc_Bot,Z_ratio,OWC,GOC,Bt,Shm_Azim,Dip_Azim,Dip_Angle
    1690,1,Alpha,70,500,2380,0.5,0,0,,,,
    2380,2,Beta,70,1600,2610,0.5,0,0,,,,
    2900,3,Gamma,70,2450,2950,0.5,2499,0,,90,,
    2950,4,Echo,60,2550,3400,0.66,3100,3050,,90,0,0
    3250,5,Delta,65,2750,3300,0.33,2950,2900,,85,2,3"""
    
    f0 = pd.read_csv("testforms.csv")
    print(f0)

    #f0 = pd.read_csv(io.StringIO(form_md), sep=",")
    cfd = extract_depth_dict(convert_df_tvd(f0,well))
    
    new_md = """topTVD,StratiNum,Formation_Name,GR_Cut,Struc_Top,Struc_Bot,Z_ratio,OWC,GOC,Bt,Shm_Azim,Dip_Azim,Dip_Angle
    1750,1,Alpha,70,500,2380,0.5,0,0,,,,
    2124,2,Beta,70,1600,2610,0.5,0,0,,,,
    2900,3,Gamma,70,2450,2950,0.5,2499,0,,90,,
    3150,4,Echo,60,2550,3400,0.66,3100,3050,,90,0,0
    4200,5,Delta,65,2750,3300,0.33,2950,2900,,85,2,3"""
    
    f1 = pd.read_csv("testforms2.csv")
    print(f1)
    
    #f1 = pd.read_csv(io.StringIO(new_md), sep=",")
    #newblankwell=create_blank_well(kb=20,gl=10,spacing=15,maxangle=90,kop=4000,azimuth=90)
    
    new_well,stringlas,devdas = get_analog(well,f0,f1,kb=20,gl=-200,ma=90,kop=1000,azi=90,debug=False)
    freshnew = getwelldev(string_las=stringlas.read(),deva=devdas,step=5)
    print(freshnew)
    print(new_well)

    output = compute_geomech(freshnew,sfs=30, tango=4000, writeFile=False,)
    outdf = output[0]
    outdf.to_csv("testout.csv")
    print(new_well.df())
    stringlas.seek(0)
    print(stringlas.read(2500))
    """
    lulu = io.StringIO(output[1])
    lulu.seek(0)
    print(lulu.read(2500))
    """



"""
Copyright (c) 2024-2025 ROCK LAB PRIVATE LIMITED
This file is part of "Stresslog" project and is released under the 
GNU Affero General Public License v3.0 (AGPL-3.0)
See the GNU Affero General Public License for more details: <https://www.gnu.org/licenses/agpl-3.0.html>
"""

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

# Vectors
x = [2, 1, 4]
y = [1, 3, 5]
z = [1, 2, 3]

def plotVectors(x, y, z, m,lx,ly,lz):
    
    #x += [2, 0, 0]
    #y += [0, 2, 0]
    #z += [0, 0, 2]

    
    fig = plt.figure()
    ax = fig.add_subplot(projection='3d')

    # Original colors and labels for each vector
    original_colors = ['r', 'g', 'b']
    original_labels = [
        f"{round(lx)}MPa", 
        f"{round(ly)}MPa", 
        f"{round(lz)}MPa"
    ]

    # Plotting the original vectors
    for i in range(len(x)):
        ax.quiver(0, 0, 0, x[i], y[i], z[i], color=original_colors[i], arrow_length_ratio=0.1, linewidth=1)
        ax.text(x[i], y[i], z[i], original_labels[i], color=original_colors[i])

    # Unit vectors (N, E, D)
    unit_vectors = [(0.5, 0, 0), (0,0.5, 0), (0, 0, 0.5)]
    unit_colors = ['k', 'k', 'k']  # 'k' for black for the unit vectors
    unit_labels = ["N", "E", "D"]
    
    # Plotting the original vectors
    for i in range(len(x)):
        ax.quiver(0, 0, 0, -x[i], -y[i], -z[i], color=original_colors[i], arrow_length_ratio=0.1, linewidth=1)
        ax.text(x[i], y[i], z[i], original_labels[i], color=original_colors[i])

    # Unit vectors (N, E, D)
    unit_vectors = [(0.5, 0, 0), (0,0.5, 0), (0, 0, 0.5)]
    unit_colors = ['k', 'k', 'k']  # 'k' for black for the unit vectors
    unit_labels = ["N", "E", "D"]

    # Plotting the unit vectors
    for i, (ux, uy, uz) in enumerate(unit_vectors):
        ax.quiver(0, 0, 0, ux, uy, uz, color=unit_colors[i], arrow_length_ratio=0.1, linewidth=1)
        ax.text(ux, uy, uz, unit_labels[i], color=unit_colors[i])

    ax.set_xlim([-m, m])
    ax.set_ylim([-m, m])
    ax.set_zlim([-m, m])
    ax.set_xlabel('N axis')
    ax.set_ylabel('E axis')
    ax.set_zlabel('D axis')
    
    unit_vectors = [(-0.5, 0, 0), (0,-0.5, 0), (0, 0, -0.5)]
    unit_colors = ['k', 'k', 'k']  # 'k' for black for the unit vectors
    unit_labels = ["S", "W", "U"]
    
    # Plotting the unit vectors
    for i, (ux, uy, uz) in enumerate(unit_vectors):
        ax.quiver(0, 0, 0, ux, uy, uz, color=unit_colors[i], arrow_length_ratio=0.1, linewidth=1)
        ax.text(ux, uy, uz, unit_labels[i], color=unit_colors[i])

    ax.set_xlim([-m, m])
    ax.set_ylim([-m, m])
    ax.set_zlim([-m, m])
    ax.set_xlabel('N axis')
    ax.set_ylabel('E axis')
    ax.set_zlabel('D axis')
    
    
    ax.set_box_aspect([1,1,1])
    ax.view_init(elev=45, azim=45)
    #plt.savefig(path, dpi=600)
    #plt.show()
    return plt

# Example usage
#plotVectors(x, y, z, 2, 23,30,40).show()
def showvec(x, y, z, m,lx,ly,lz,path):
    plotVectors(x, y, z, m,lx,ly,lz).show()
    return
def savevec(x, y, z, m,lx,ly,lz,path):
    plt = plotVectors(x, y, z, m,lx,ly,lz)
    plt.savefig(path, dpi=600)
    plt.close()
    return



"""
Copyright (c) 2024-2025 ROCK LAB PRIVATE LIMITED
This file is part of "Stresslog" project and is released under the 
GNU Affero General Public License v3.0 (AGPL-3.0)
See the GNU Affero General Public License for more details: <https://www.gnu.org/licenses/agpl-3.0.html>
"""
import warnings
warnings.filterwarnings("ignore")

import numpy as np
import pandas as pd
import math
from matplotlib import pyplot as plt
import random
from io import StringIO

from .thirdparty import datasets_to_las

# Option 1: Using set_printoptions to control decimal places
np.set_printoptions(precision=3, suppress=True)  # suppress will prevent scientific notation



def random_curves_dataframe(
    start_depth, stop_depth, step,
    matrick=60, mudline=210, dt_ncts=0.0008,
    rhoappg=17, a=0.08, water=8.6, mudweight=9.8, glwd=-1, seed=None, drop=[],to=0
):
    """Generate synthetic well log data with random variations.

    Parameters
    ----------
    start_depth : float
        Starting depth for the well log in meters
    stop_depth : float
        Ending depth for the well log in meters
    step : float
        Depth increment between measurements in meters
    matrick : float, optional
        Matrix sonic transit time in us/ft, by default 60
    mudline : float, optional
        Mudline sonic transit time in us/ft, by default 210
    dt_ncts : float, optional
        Compaction trend coefficient, by default 0.0008
    rhoappg : float, optional
        Apparent density gradient in ppg, by default 17
    a : float, optional
        Density compaction coefficient, by default 0.08
    water : float, optional
        Water density in ppg, by default 8.6
    mudweight : float, optional
        Mud weight in ppg, by default 9.8
    glwd : float, optional
        Ground level water depth in meters, by default -1
    seed : int, optional
        Random seed for reproducibility, by default None
    drop : list, optional
        List of curve names to exclude from output, by default []

    Returns
    -------
    pandas.DataFrame
        DataFrame containing synthetic well log curves including:
        - DEPT: Depth
        - DTCO: Compressional sonic
        - RHOB: Bulk density
        - GR: Gamma ray
        - NU: Poisson's ratio
        - DTS: Shear sonic
        - NPHI: Neutron porosity
        - ILD: Induction log deep resistivity
    """
    if seed is None:
        seed = random.randint(0, 1000000)
        print("Seed: ", seed)
    dtco_seed = gr_seed = nu_seed = rhob_seed = nphi_seed = ild_seed = seed

    # Set up random generators
    dtco_rng = np.random.default_rng(dtco_seed)
    gr_rng = np.random.default_rng(gr_seed)
    nu_rng = np.random.default_rng(nu_seed)
    rhob_rng = np.random.default_rng(rhob_seed)
    nphi_rng = np.random.default_rng(nphi_seed)
    ild_rng = np.random.default_rng(ild_seed)

    # Generate DEPT column
    DEPT = np.arange(start_depth, stop_depth, step)

    # Calculate DTCO with controlled randomness
    DTCO = (
        matrick
        + (mudline - matrick) * np.exp(-dt_ncts * DEPT)
        + np.clip(np.cumsum(dtco_rng.normal(0, 5, len(DEPT))) * 0.1, -50, 50)
    )

    # Calculate RHOB
    rhoppg = np.zeros_like(DEPT)
    rhogcc = np.zeros_like(DEPT)

    for i in range(len(DEPT)):
        tvdbgl = DEPT[i] - glwd
        tvdmsl = DEPT[i]

        if glwd < 0:
            if tvdbgl >= 0:
                rhoppg[i] = rhoappg + ((DEPT[i] / 3125) ** a)
                rhogcc[i] = (0.11982642731 * rhoppg[i]) #+ np.clip(np.cumsum(rhob_rng.normal(0, 5, 1)) * 0.005, -0.09, 0.09)
            else:
                if tvdmsl < 0:
                    rhoppg[i] = 8.34540426515252 * water
                    rhogcc[i] = 0.11982642731 * rhoppg[i] 
                else:
                    rhoppg[i] = 0
                    rhogcc[i] = 0
        else:
            if tvdbgl >= 0:
                rhoppg[i] = rhoappg + ((tvdbgl / 3125) ** a)
                rhogcc[i] = (0.11982642731 * rhoppg[i]) #+ np.clip(np.cumsum(rhob_rng.normal(0, 5, 1)) * 0.005, -0.09, 0.09)
            else:
                rhoppg[i] = 0
                rhogcc[i] = 0

    RHOB = rhogcc + np.clip(np.cumsum(rhob_rng.normal(0, 5, len(DEPT))) * 0.000009, -0.009, 0.009)

    # Generate GR with controlled randomness
    GR = np.clip(np.cumsum(gr_rng.normal(0, 5, len(DEPT))) * 0.1, 0, 150)

    # Generate NU with controlled randomness
    NU = np.clip(nphi_rng.normal(0.25, 0.075, len(DEPT)), 0.15, 0.45)#np.clip(np.cumsum(nu_rng.normal(0, 5, len(DEPT))) * 0.00085, 0.2, 0.40)

    # Derived columns: Vp, Vs, and DTS
    Vp = 1 / DTCO
    Vs = np.sqrt(((Vp ** 2) * (2 * NU - 1)) / (2 * (NU - 1)))
    DTS = 1 / Vs
    
    # Generate NPHI and ILD with controlled randomness
    NPHI = np.clip(0.5 - nphi_rng.normal(0, 0.05, len(DEPT)), 0, 1)  # Normalized porosity
    ILD = np.clip(10 ** ild_rng.normal(1, 0.2, len(DEPT)), 0.02, 200)  # Log-normal resistivity

    # Create DataFrame
    data = {
        "DEPT": DEPT,
        "DTCO": DTCO,
        "RHOB": RHOB,
        "GR": GR,
        "NU": NU,
        "DTS": DTS,
        "NPHI": NPHI,
        "ILD": ILD
    }
    # Drop specified curves
    data = {key: value for key, value in data.items() if key not in drop}
    df = pd.DataFrame(data)
    mask = df["DEPT"] <= to
    df.loc[mask, df.columns.difference(["DEPT"])] = np.nan

    return df



def create_header(name, uwi, strt, stop, step, 
                     comp=None, fld=None, loc=None, cnty=None, stat=None, ctry=None, 
                     api=None, date=None, srvc=None, lati=None, long=None, gdat=None, kb = "1", gl = "0", nully="-999.25"):
    """Create a LAS file header with well information.

    Parameters
    ----------
    name : str
        Well name
    uwi : str
        Unique Well Identifier
    strt : str
        Start depth
    stop : str
        Stop depth
    step : str
        Depth step size
    comp : str, optional
        Company name
    fld : str, optional
        Field name
    loc : str, optional
        Location description
    cnty : str, optional
        County name
    stat : str, optional
        State name
    ctry : str, optional
        Country name
    api : str, optional
        API number
    date : str, optional
        Log date (DD-MMM-YYYY format)
    srvc : str, optional
        Service company name
    lati : str, optional
        Latitude (in degrees)
    long : str, optional
        Longitude (in degrees)
    gdat : str, optional
        Geodetic datum
    kb : str, optional
        Kelly bushing elevation in meters, by default "1"
    gl : str, optional
        Ground level elevation in meters, by default "0"
    nully : str, optional
        Null value indicator, by default "-999.25"

    Returns
    -------
    pandas.DataFrame
        DataFrame containing formatted LAS file header information
    """
    # Core mandatory content
    csv_parts = [
        """\
original_mnemonic,mnemonic,unit,value,descr,section
WRAP,WRAP,,NO,One Line per depth step,Version
PROD,PROD,,RockLab,LAS Producer,Version
PROG,PROG,,Stresslog1.4.x,LAS Program name and version,Version
CREA,CREA,,2023/07/13 12:42                          :LAS Creation date {{YYYY/MM/DD hh,mm}},Version
"""
    ]
    
    # Conditional content based on optional parameters

    if strt:
        csv_parts.append(f"STRT,STRT,M,{strt},START DEPTH,Well")
    if stop:
        csv_parts.append(f"STOP,STOP,M,{stop},STOP DEPTH,Well")
    if step:
        csv_parts.append(f"STEP,STEP,M,{step},STEP,Well")
    if comp:
        csv_parts.append(f"COMP,COMP, ,{comp},COMPANY,Well")
    if name:
        csv_parts.append(f"WELL,WELL, ,{name},WELL NAME,Well")
    if uwi:
        csv_parts.append(f"UWI,UWI, ,{uwi},WELL UWI,Well")
    if fld:
        csv_parts.append(f"FLD,FLD, ,{fld},FIELD,Well")
    if loc:
        csv_parts.append(f"LOC,LOC, ,{loc},LOCATION,Well")
    if cnty:
        csv_parts.append(f"CNTY,CNTY, ,{cnty},COUNTY,Well")
    if stat:
        csv_parts.append(f"STAT,STAT, ,{stat},STATE,Well")
    if ctry:
        csv_parts.append(f"CTRY,CTRY, ,{ctry},COUNTRY,Well")
    if api:
        csv_parts.append(f"API,API, , ,API NUMBER,Well")
    if date:
        csv_parts.append(f"DATE,DATE, ,{date},LOG DATE {{DD-MMM-YYYY}},Well")
    if srvc:
        csv_parts.append(f"SRVC,SRVC, ,{srvc},SERVICE COMPANY,Well")
    if lati:
        csv_parts.append(f"LATI,LATI,DEG,\"{lati}\",LATITUDE,Well")
    if long:
        csv_parts.append(f"LONG,LONG,DEG,\"{long}\",LONGITUDE,Well")
    if gdat:
        csv_parts.append(f"GDAT,GDAT, ,{gdat},GeoDetic Datum,Well")
    if kb:
        csv_parts.append(f"EKB,GDAT,m ,{kb},Elevation of Kelly Bushing,Well")
        csv_parts.append(f"KB,GDAT,m ,{kb},Elevation of Kelly Bushing,Well")
    if gl:
        csv_parts.append(f"EGL,GDAT,m ,{gl},Elevation of Ground Level,Well")
        csv_parts.append(f"GL,GDAT,m ,{gl},Elevation of Ground Level,Well")
    
    
    # Static curve information
    csv_parts.append("""\
DEPT,DEPT,M,,DEPTH (BOREHOLE),Curves
DTCO,DTCO,US/F,,Delta-T Compressional,Curves
RHOB,RHOB,G/CC,,Bulk Density,Curves
GR,GR,GAPI,,Gamma Ray,Curves
NU,NU, ,,POISSON RATIO,Curves
DTS,DTS,US/F,,Delta-T Shear,Curves""")

    # Join all parts into one CSV string
    csv_content = "\n".join(csv_parts)
    #print(csv_content)
    # Create DataFrame from the CSV string
    df = pd.read_csv(StringIO(csv_content))
    return df



def create_random_las(
    starter=0, 
    stopper=6000, 
    stepper=0.15, 
    key=None, 
    well_name="WonderWell", 
    company="Wonder", 
    field="Brown", 
    location="Morrdor", 
    county="Shire", 
    latitude="60° 34' 35.15\" N", 
    longitude="3° 26' 36.15\" E", 
    geodetic_datum="WGS84",
    writeFile=False,
    debug=False,
    kb=2,
    gl=0,
    drop=[],
):
    """
    Process well log data by generating a DataFrame, creating a header, plotting, and optionally writing a LAS file.
    
    Parameters:
    -----------
    starter : float, optional
        Starting depth of the well log. Default is 0.
    stopper : float, optional
        Stopping depth of the well log. Default is 2800.
    stepper : float, optional
        Step size for depth intervals. Default is 0.15.
    key : int, optional
        Random seed for data generation. If None, a random seed will be generated.
    well_name : str, optional
        Name of the well. Default is "WonderWell".
    company : str, optional
        Company name. Default is "Wonder".
    field : str, optional
        Field name. Default is "Brown".
    location : str, optional
        Location name. Default is "Morrdor".
    county : str, optional
        County name. Default is "Shire".
    latitude : str, optional
        Latitude of the well. Default is "60° 34' 35.15\" N".
    longitude : str, optional
        Longitude of the well. Default is "3° 26' 36.15\" E".
    geodetic_datum : str, optional
        Geodetic datum. Default is "WGS84".
    writeFile : bool, optional
        If True, writes LAS file to disk. If False, returns string buffer. Default is True.
    
    Returns:
    --------
    If writeFile is True:
        tuple: (DataFrame with well log data, Header DataFrame, Filename of LAS file)
    If writeFile is False:
        tuple: (DataFrame with well log data, Header DataFrame, StringIO buffer of LAS file)
    """
    # Set random seed if not provided
    if key is None:
        key = random.randint(0, 1000)
    if gl<0:
        topoffset=(-gl)+kb
    else:
        topoffset = kb-gl
    df = random_curves_dataframe(starter, stopper, stepper, seed=key, drop=drop, to=topoffset)
    
    # Create header DataFrame
    hdf = create_header(
        name=f"{well_name}-{key}", 
        uwi=f"WELL#{key}", 
        strt=str(starter), 
        stop=str(stopper), 
        step=str(stepper), 
        comp=company, 
        fld=field, 
        loc=location, 
        cnty=county,
        lati=latitude, 
        long=longitude, 
        gdat=geodetic_datum,
        kb=kb,
        gl=gl,
    )
    if writeFile or debug:
        # Determine columns to plot (excluding DEPT)
        columns_to_plot = [col for col in df.columns if col != 'DEPT']
        
        # Columns to reverse x-axis for
        reverse_x_columns = ['DTS', 'DTCO', 'NPHI']
        
        # Calculate subplot layout
        num_cols = len(columns_to_plot)
        
        with plt.xkcd():
            # Create side-by-side subplots
            plt.figure(figsize=(2 * num_cols, 6))
            
            for i, col in enumerate(columns_to_plot, 1):
                plt.subplot(1, num_cols, i)
                
                # Plot the data
                plt.plot(df[col], df['DEPT'])
                
                # Invert y-axis
                plt.gca().invert_yaxis()
                
                # Conditionally reverse x-axis for specific columns
                if col in reverse_x_columns:
                    plt.gca().invert_xaxis()
                
                plt.xlabel(col)
                plt.ylabel(None)
                plt.title(f"{col}")
            
            plt.tight_layout()
            
            plt.show()
    
    # Convert to LAS file
    las_string = datasets_to_las(None, {'Curves': df, 'Header': hdf}, {'DEPT': 'm', 'DTCO': 'us/f', 'RHOB': 'g/cc', 'GR': 'gAPI', 'DTS': 'us/f'})
    
    if writeFile:
        # Write LAS file to disk
        filename = f"randomwonder{key}.las"
        with open(filename, 'w') as file:
            file.write(las_string)
        print(f"LAS file has been written to {filename}")
        return df, hdf, filename
    else:
        # Return string buffer instead of writing to file
        string_buffer = StringIO(las_string)
        return df, hdf, string_buffer
    
from welly import Well

def getwelldev(string_las=None, wella=None, deva=None, kickoffpoint=None, final_angle=None, rateofbuild=None, azimuth=None, step=None):
    """Calculate well deviation data and update well object with TVD information.

    Parameters
    ----------
    string_las : str, optional
        LAS file content as string
    wella : welly.Well, optional
        Well object to process
    deva : pandas.DataFrame, optional
        Existing deviation data with columns 'MD', 'INC', 'AZIM'
    kickoffpoint : float, optional
        Depth at which well deviation begins
    final_angle : float, optional
        Target inclination angle in degrees
    rateofbuild : float, optional
        Rate of angle build in degrees per meter
    azimuth : float, optional
        Well azimuth in degrees

    Returns
    -------
    welly.Well
        Updated well object with deviation data and TVD calculations

    Notes
    -----
    Either provide string_las or wella, and either deva or kickoffpoint+final_angle+rateofbuild+azimuth
    """
    if wella is None:
        wella = Well.from_las(string_las, index="m")
    
    depth_track = wella.df().index

    if deva is None and kickoffpoint is not None and final_angle is not None and rateofbuild is not None and azimuth is not None:
        if final_angle>=90:
            final_angle=89.999
        start_depth = depth_track[0]
        final_depth = depth_track[-1]
        spacing = depth_track[-3] - depth_track[-4]
        if step is not None:
            spacing=step
        # Create a depth array with the same spacing
        md = np.arange(start_depth, final_depth + spacing, spacing)
        inc = np.zeros_like(md)
        azm = np.full_like(md, azimuth, dtype=float)

        # Calculate inclination from kickoff point onwards
        for i, depth in enumerate(md):
            if depth >= kickoffpoint:
                increment = (depth - kickoffpoint) * rateofbuild
                inc[i] = min(increment, final_angle)  # Cap inclination at final_angle

        deva = pd.DataFrame({"MD": md, "INC": inc, "AZIM": azm})

    if deva is not None:
        start_depth = depth_track[0]
        spacing = depth_track[-3] - depth_track[-4]
        if step is not None:
            spacing=step
        padlength = int(start_depth / spacing)
        padval = np.zeros(padlength)

        for i in range(1, padlength):
            padval[-i] = start_depth - (spacing * i)

        md = np.append(padval, depth_track)
        mda = pd.to_numeric(deva.iloc[:, 0], errors="coerce")
        inca = pd.to_numeric(deva.iloc[:, 1], errors="coerce")
        azma = pd.to_numeric(deva.iloc[:, 2], errors="coerce")
        inc = np.interp(md, mda, inca)
        azm = np.interp(md, mda, azma)
    else:
        start_depth = depth_track[0]
        spacing = (depth_track[-3] - depth_track[-4])
        if step is not None:
            spacing=step
        padlength = int(start_depth / spacing)
        padval = np.zeros(padlength)

        for i in range(1, padlength):
            padval[-i] = start_depth - (spacing * i)

        md = np.append(padval, depth_track)
        inc = np.zeros(len(depth_track) + padlength)
        azm = np.zeros(len(depth_track) + padlength)

    dz = np.transpose([md, inc, azm])
    dz = pd.DataFrame(dz).dropna()

    wella.location.add_deviation(dz, wella.location.td)
    tvdg = wella.location.tvd
    #tvdg[-1] = tvdg[-2]+(tvdg[-2]-tvdg[-3])
    md = wella.location.md
    #md[-1] = md[-2]+(md[-2]-md[-3])

    from welly import curve
    MD = curve.Curve(md, mnemonic='MD', units='m', index=md)
    wella.data['MD'] = MD
    TVDM = curve.Curve(tvdg, mnemonic='TVDM', units='m', index=md)
    wella.data['TVDM'] = TVDM

    wella.unify_basis(keys=None, alias=None, step=spacing)

    #print("Great Success!! :D")
    return wella

def create_random_well(kb,gl,kop=0, maxangle=0, step=0.15,starter=0,stopper=5500, drop=None,seed=None):
    """Create a random well object with specified parameters.

    Parameters
    ----------
    kb : float
        Kelly bushing elevation in meters
    gl : float
        Ground level elevation in meters
    kop : float, optional
        Kickoff point depth in meters, by default 0
    maxangle : float, optional
        Maximum deviation angle in degrees, by default 0
    step : float, optional
        Depth step size in meters, by default 0.15
    drop : list, optional
        List of curves to exclude from output, by default []

    Returns
    -------
    welly.Well
        Well object containing randomly generated log data
    """
    if drop is None:
        drop = []
    if seed is None:
        data_frame_2, header_frame_2, string_buffer = create_random_las(kb=kb,gl=gl, stepper=step, stopper=stopper, drop=drop)
        string_data = string_buffer.getvalue()
        return Well.from_las(string_data)
    else:
        data_frame_2, header_frame_2, string_buffer = create_random_las(kb=kb,gl=gl, stepper=step, stopper=stopper, drop=drop, key=seed)
        string_data = string_buffer.getvalue()
        return Well.from_las(string_data)


import dlisio
dlisio.common.set_encodings(['utf-8','latin-1'])
from dlisio import dlis

def get_dlis_header(path):
    """
    Extract header information from a DLIS file.

    Parameters
    ----------
    path : str
        Path to the DLIS file.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing two columns:
        - Mnemonic: Channel names from the DLIS file
        - Description: Long names/descriptions of the channels
    """
    
    # Load the DLIS file
    f, *tail = dlis.load(path)
    # Get the list of frames
    frames = f.frames
    
    chnames = []
    description = []
    for ch in f.channels:
        chnames.append(ch.name)
        description.append(ch.long_name)
    return pd.DataFrame(list(zip(chnames,description)),columns =['Mnemonic', 'Description'])


def get_dlis_data(path, aliases=None, depthunits='m', resample_interval=0.1, flatten=False, debug=False):
    """
    Extract data from a DLIS file with unit conversion and header processing.

    Parameters
    ----------
    path : str
        Path to the DLIS file.
    aliases : dict or None
        Dictionary mapping alias names to lists of possible mnemonics.
        If None, all 1-dimensional curves present in the DLIS file will be extracted.
        Multi-dimensional curves will be flattened into individual sub-curves.
    depthunits : str, optional
        Desired depth units for output. Options are 'm' (meters) or 'f' (feet).
        Default is 'm'.

    Returns
    -------
    combined_df : pandas.DataFrame
        Combined data from all frames with depth as index and specified curves as columns.
        When aliases is None, includes all 1-dimensional curves found in the DLIS file.
        Depth values are converted according to depthunits parameter.
    c_units : dict
        Dictionary mapping curve names to their units.
    header_df : pandas.DataFrame
        DataFrame containing header information.
    parameters : dict
        Dictionary containing raw parameter values from the DLIS file.
    """
    if resample_interval < 0.1469:
        resample_interval = 0.1469
    
    f, *tail = dlis.load(path)
    origin, *origin_tail = f.origins
    
    dataframes = {}
    c_units = {}
    parameters = {}
    long_names = {}
    p_units = {}
    p_section = {}
    
    conversion_factors = {
        'ft': 0.3048,
        'f': 0.3048,
        'feet': 0.3048,
        '0.1in': 0.00254,
        'inches': 0.0254,
        'in': 0.0254
    }

    for param in f.parameters:
        try:
            if hasattr(param, 'values') and param.values is not None:
                if len(param.values) == 1:
                    value = str(param.values[0])
                else:
                    value = str(list(param.values))
                
                if value != "-999.25" and value != "" and value != "0" and len(value) < 100:
                    parameters[param.name] = value
                    long_names[param.name] = param.long_name
                    
                    try:
                        p_units[param.name] = param.attic['VALUES'].units
                    except (KeyError, AttributeError):
                        p_units[param.name] = ''
                        
                    p_section[param.name] = "Parameter"
        except (ValueError, AttributeError) as e:
            print(f"Warning: Skipping parameter {param.name} due to error: {str(e)}")
            continue

    header_df = pd.DataFrame([
        {
            'original_mnemonic': param,
            'mnemonic': param,
            'unit': p_units.get(param, ''),
            'value': value,
            'descr': long_names.get(param, ''),
            'section': p_section.get(param, 'Parameter')
        }
        for param, value in parameters.items()
    ])

    header_df = header_df[['original_mnemonic', 'mnemonic', 'unit', 'value', 'descr', 'section']]
    
    all_curves = set()
    skipped_curves = set()

    for frame in f.frames:
        curves = frame.curves(strict=False)
        all_curves.update(curves.dtype.names)
        
        extracted_curves = {}
        
        index_channel = frame.index
        if index_channel:
            tdep_channel = next((ch for ch in frame.channels if ch.name == index_channel), None)
            if tdep_channel:
                tdep_unit = tdep_channel.units.lower()
                conversion_factor = conversion_factors.get(tdep_unit, 1)
                index_data = curves[index_channel]
                if len(index_data.shape) == 1:
                    extracted_curves[index_channel] = index_data * conversion_factor
                    if index_channel not in c_units:
                        c_units[index_channel] = tdep_unit
                else:
                    print(f"Warning: Index channel {index_channel} is multi-dimensional and will be skipped")
                    continue

        found_useful_data = False
        if aliases is None:
            for curve_name in curves.dtype.names:
                if curve_name != index_channel:
                    curve_data = curves[curve_name]
                    if len(curve_data.shape) == 1:
                        extracted_curves[curve_name] = curve_data
                        found_useful_data = True
                        curve_channel = next((ch for ch in frame.channels if ch.name == curve_name), None)
                        if curve_channel and curve_name not in c_units:
                            c_units[curve_name] = curve_channel.units
                    else:
                        if flatten:
                            print(f"Flattening multi-dimensional curve: {curve_name} with shape {curve_data.shape}") if debug else None
                            for i in range(curve_data.shape[1]):
                                sub_curve_name = f"{curve_name}_{i+1}"
                                sub_curve_data = curve_data[:, i]
                                if len(sub_curve_data.shape) == 1:
                                    extracted_curves[sub_curve_name] = sub_curve_data
                                    found_useful_data = True
                                    curve_channel = next((ch for ch in frame.channels if ch.name == curve_name), None)
                                    if curve_channel and sub_curve_name not in c_units:
                                        c_units[sub_curve_name] = curve_channel.units
                        else:
                            print(f"Skipping multi-dimensional curve: {curve_name}") if debug else None
                            skipped_curves.add(curve_name)
        else:
            for alias, mnemonics in aliases.items():
                for mnemonic in mnemonics:
                    if mnemonic in curves.dtype.names:
                        curve_data = curves[mnemonic]
                        if len(curve_data.shape) == 1:
                            extracted_curves[mnemonic] = curve_data
                            found_useful_data = True
                            curve_channel = next((ch for ch in frame.channels if ch.name == mnemonic), None)
                            if curve_channel and mnemonic not in c_units:
                                c_units[mnemonic] = curve_channel.units
                            break
                        elif flatten:
                            print(f"Flattening multi-dimensional curve: {mnemonic} with shape {curve_data.shape}") if debug else None
                            for i in range(curve_data.shape[1]):
                                sub_curve_name = f"{mnemonic}_{i+1}"
                                sub_curve_data = curve_data[:, i]
                                if len(sub_curve_data.shape) == 1:
                                    extracted_curves[sub_curve_name] = sub_curve_data
                                    found_useful_data = True
                                    curve_channel = next((ch for ch in frame.channels if ch.name == mnemonic), None)
                                    if curve_channel and sub_curve_name not in c_units:
                                        c_units[sub_curve_name] = curve_channel.units
                            break
                        else:
                            print(f"Skipping multi-dimensional curve: {mnemonic}") if debug else None
                            skipped_curves.add(mnemonic)
                            continue

        if found_useful_data:
            try:
                frame_df = pd.DataFrame(extracted_curves)
                if index_channel and index_channel in frame_df.columns:
                    frame_df.set_index(index_channel, inplace=True)
                if not frame_df.empty:
                    dataframes[frame.name] = frame_df
            except ValueError as e:
                print(f"Warning: Error creating DataFrame for frame {frame.name}: {str(e)}")
                continue

    if debug:
        print("All encountered curves:")
        print(", ".join(sorted(all_curves)))
        if skipped_curves:
            print("\nSkipped multi-dimensional curves:")
            print(", ".join(sorted(skipped_curves)))

    if not dataframes:
        raise ValueError("No valid data frames could be created from the DLIS file")
    
    combined_df = pd.concat(dataframes.values(), axis=1, join='outer').sort_index()
    # Remove duplicate columns after initial concatenation
    combined_df = combined_df.loc[:, ~combined_df.columns.duplicated(keep='first')]
    combined_df.replace(-999.25, float("nan"), inplace=True)
    
    if depthunits == 'm':
        combined_df.index = combined_df.index * 0.00245
    elif depthunits == 'f':
        combined_df.index = combined_df.index * 0.00833
    combined_df.sort_index(inplace=True)
    c_units[index_channel] = depthunits
    
    if resample_interval is not None:
        min_depth = combined_df.index.min()
        max_depth = combined_df.index.max()

        start = np.floor(min_depth / resample_interval) * resample_interval
        end = np.ceil(max_depth / resample_interval) * resample_interval
        bins = np.arange(start, end + resample_interval, resample_interval)
        labels = bins[:-1]

        combined_df['depth_bin'] = pd.cut(
            combined_df.index, 
            bins=bins, 
            labels=labels, 
            include_lowest=True, 
            right=False
        )
        resampled_df = combined_df.groupby('depth_bin', observed=False).mean()
        resampled_df = resampled_df.reindex(labels)
        resampled_df.index.name = combined_df.index.name
        resampled_df.sort_index(inplace=True)
        combined_df = resampled_df

    aux_header = {
        'STRT': combined_df.index.min(),
        'STOP': combined_df.index.max(),
        'STEP': np.mean(np.diff(combined_df.index)),
        'NULL': -999.25,
        'UWI': origin.well_id,
        'WELL': origin.well_name,
        'SRVC': origin.producer_name,
        'COMP': origin.company,
        'FLD': origin.field_name
    }
    
    aux_units = {
        'STRT': 'M',
        'STOP': 'M',
        'STEP': 'M',
        'NULL': '',
        'UWI': '',
        'WELL': '',
        'SRVC': '',
        'COMP': '',
        'FLD': ''
    }

    aux_df = pd.DataFrame([
        {
            'original_mnemonic': k,
            'mnemonic': k,
            'unit': aux_units[k],
            'value': v,
            'descr': '',
            'section': 'Well'
        }
        for k, v in aux_header.items()
    ])

    header_df = pd.concat([aux_df, header_df], ignore_index=True)
    
    index_name = combined_df.index.name or 'DEPTH'
    combined_df = combined_df.copy()
    combined_df[index_name] = combined_df.index
    columns = combined_df.columns.tolist()
    columns.insert(0, columns.pop(columns.index(index_name)))
    combined_df = combined_df[columns]
    combined_df.index.name = index_name

    # Final deduplication of columns after all processing
    combined_df = combined_df.loc[:, ~combined_df.columns.duplicated(keep='first')]
    # Update c_units to include only remaining columns
    existing_columns = combined_df.columns
    c_units = {k: v for k, v in c_units.items() if k in existing_columns}

    return combined_df, c_units, header_df, parameters


def get_las_from_dlis(path,aliases,depthunit='m',step=0.15, debug=False):
    """
    DEPRECATED in favor of get_well_from_dlis
    """
    y = get_dlis_data(path,aliases,resample_interval=step, debug=debug)
    return datasets_to_las(None, {'Curves': y[0], 'Header': y[2]}, y[1])

def get_well_from_dlis(path,aliases=None,depthunit='m',step=0.15, debug=False):
    """
    Reads a dlis file using dlisio and converts the data first to a las string and then to a welly.Well object.

    Parameters
    ----------
    path : str
        Path to the DLIS file.
    aliases: dict
        The dict of mnemonics used to decide which of the available dlis curves will be imported. If None (default) then all the curves are included.
    depthunit: str
        String indicating the desired depth unit of the parsed well. Pass 'f' for depth in feet, or 'm' (default) for depth in metres.
    step: float
        Desired sample spacing of the parsed well data, in the unit chosen. Do not try to oversample much, results in weird aliasing artefacts if appreciably below the actual sample spacing.
    
    Returns
    -------
    welly.Well
        A well object representing the dlis file, resampled according to the given step.
    """
    y = get_dlis_data(path,aliases,resample_interval=step, depthunits=depthunit, debug=debug)
    z = datasets_to_las(None, {'Curves': y[0], 'Header': y[2]}, y[1])
    well = Well.from_las(z, index=depthunit)
    start = well.header.loc[well.header['mnemonic'] == 'STRT', 'value'].values[0]
    stop = well.header.loc[well.header['mnemonic'] == 'STOP', 'value'].values[0]
    manual_basis = np.arange(start, stop, step)
    well.unify_basis(basis=manual_basis)
    return well

# Example usage
if __name__ == "__main__":
    # Demonstrate the function with file writing
    #data_frame, header_frame, filename = process_well_data(writeFile=True)
    
    # Demonstrate the function with string buffer
    data_frame_2, header_frame_2, string_buffer = create_random_las(kb=50,gl=-30, debug=True)
    
    # Optional: read string buffer content
    if not isinstance(string_buffer, str):
        print("\nLAS File Content (first 1500 chars):")
        string_buffer.seek(0)
        print(string_buffer.read(1500))
    x = create_random_well(kb=35,gl=20)
    x = getwelldev(wella = x)#,kickoffpoint=1000,final_angle=85, azimuth=89, rateofbuild=0.2)
    print(x)
    y=x.location.trajectory()
    z = x.location.position
    # Assuming z is your array
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')

    # Plot 3D line
    ax.plot3D(z[:, 0], z[:, 1], -z[:, 2])

    # Labels
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Depth (Z)')

    plt.show()




import os
from PIL import Image

def reduce_png_resolution(folder: str):
    EXCEPTIONS = {'WellPlot.png', 'overlay.png', 'FMI-2630.png'}
    output_folder = os.path.join(folder, "resized")
    os.makedirs(output_folder, exist_ok=True)
    
    if not os.path.exists(folder):
        print(f"Error: Folder '{folder}' does not exist.")
        return
    
    for filename in os.listdir(folder):
        if filename.lower().endswith('.png') and filename not in EXCEPTIONS:
            file_path = os.path.join(folder, filename)
            output_path = os.path.join(output_folder, filename)
            
            user_input = input(f"Resize {filename}? (y/N): ").strip().lower()
            if user_input != 'y':
                print(f"Skipping {filename}.")
                continue
            
            scale_factor = input("Enter the scale factor (default: 10): ")
            scale_factor = int(scale_factor) if scale_factor.isdigit() else 10
            
            try:
                with Image.open(file_path) as img:
                    new_width = max(img.width // scale_factor, 1)
                    new_height = max(img.height // scale_factor, 1)
                    img = img.resize((new_width, new_height), Image.LANCZOS)
                    img.save(output_path)
                print(f"Reduced resolution of {filename} from ({img.width}, {img.height}) to ({new_width}, {new_height}).")
            except Exception as e:
                print(f"Failed to process {filename}: {e}")

if __name__ == "__main__":
    folder = input("Enter the path to the folder: ")
    reduce_png_resolution(folder)


"""
Copyright (c) 2024-2025 ROCK LAB PRIVATE LIMITED
This file is part of "Stresslog" project and is released under the 
GNU Affero General Public License v3.0 (AGPL-3.0)
See the GNU Affero General Public License for more details: <https://www.gnu.org/licenses/agpl-3.0.html>
"""

# BoreStab.py
from .BoreStab import (
    get_principal_stress,
    getVertical,
    getAlignedStress,
    getRota,
    getStens,
    getStrikeDip,
    getEuler,
    getOrit,
    getSigmaTT,
    getHoop,
    draw
)

from .thirdparty import datasets_to_las

from .syntheticLogs import create_random_las, getwelldev, create_random_well, create_header, get_las_from_dlis, get_well_from_dlis, get_dlis_data, get_dlis_header

# DrawSP.py
from .DrawSP import drawSP, getSP

# failure_criteria.py
from .failure_criteria import (
    mod_lad_cmw,
    mod_lad_cmw2,
    mogi_failure,
    mohr_failure,
    lade_failure,
    mogi,
    lade,
    zhang_sanding_cwf,
    willson_sanding_cwf,
    plot_sanding
)

from .forecast import convert_df_tvd, get_analog

# geomechanics.py
from .geomechanics import compute_geomech, remove_curves, add_curves, find_TVD, find_nearest_depth

# hydraulics.py
from .hydraulics import (
    getColumnHeights,
    getPPfromTop,
    getPPfromTopRecursive,
    getGasDensity,
    getHydrostaticPsi,
    compute_optimal_offset,
    compute_optimal_gradient
)

# obgppshmin.py
from .obgppshmin import (
    get_OBG_pascals_vec,
    get_PPgrad_Zhang_gcc,
    get_PP_grad_Zhang_gcc_vec,
    get_PPgrad_Eaton_gcc,
    get_PPgrad_Eaton_gcc_vec,
    get_PPgrad_Dxc_gcc,
    get_PPgrad_Dxc_gcc_vec,
    get_Dxc,
    get_Dxc_vec,
    get_Shmin_grad_Daine_ppg,
    get_Shmin_grad_Daine_ppg_vec
)

# forecast.py
from .forecast import get_analog, convert_df_tvd, create_blank_well

# plotangle.py
from .plotangle import plotfrac

from .unit_converter import convert_rop, convert_wob, convert_ecd, convert_torque, convert_flowrate

# Plotter.py
from .Plotter import plot_logs_labels

# We want to define __all__ to control what gets imported with "from package import *"
__all__ = [
    # BoreStab
    "get_principal_stress", "getVertical", "getAlignedStress", "getRota", "getStens",
    "getStrikeDip", "getEuler", "getOrit", "getSigmaTT", "getHoop", "draw",
    # DrawSP
    "drawSP", "getSP",
    # failure_criteria
    "mod_lad_cmw", "mod_lad_cmw2", "mogi_failure", "mohr_failure", "lade_failure",
    "mogi", "lade", "zhang_sanding_cwf", "willson_sanding_cwf", "plot_sanding",

    # geomechanics
    "compute_geomech",

    # hydraulics
    "getColumnHeights", "getPPfromTop", "getPPfromTopRecursive",
    "getGasDensity", "getHydrostaticPsi", "compute_optimal_offset", "compute_optimal_gradient",
    # obgppshmin
    "get_OBG_pascals_vec", "get_PPgrad_Zhang_gcc", "get_PP_grad_Zhang_gcc_vec",
    "get_PPgrad_Eaton_gcc", "get_PPgrad_Eaton_gcc_vec", "get_PPgrad_Dxc_gcc",
    "get_PPgrad_Dxc_gcc_vec", "get_Dxc", "get_Dxc_vec", "get_Shmin_grad_Daine_ppg",
    "get_Shmin_grad_Daine_ppg_vec",
    # plotangle
    "plotfrac",
    # Misc.
    "getwelldev","create_random_well","find_TVD","add_curves","remove_curves","create_header","datasets_to_las","get_las_from_dlis","get_dlis_data", "get_dlis_header",
    # forecast
    "get_analog", "convert_df_tvd",
    # Plotter
    "plot_logs_labels"
]


import pathlib
import setuptools

setuptools.setup(
    name="stresslog",
    version="1.6.15",
    description="Library for stress calculations from well logs",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    url="https://www.rocklab.in/stresslog",
    author="Arkadeep Ghosh",
    author_email="arkadeep_ghosh@rocklab.in",
    license="AGPL-3.0",
    project_urls={
        "Documentation":"https://stresslog.readthedocs.io/en/stable/",
        "Source":"https://github.com/GeoArkadeep/Stresslog"
    },
    classifiers=[
        "Development Status :: 6 - Mature",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Scientific/Engineering",
    ],
    python_requires=">=3.9,<=3.14",
    install_requires=["pandas>=2.0","numpy","scipy","welly","pint","matplotlib", "plotly","dlisio","setuptools>=42,<81"],
    packages=setuptools.find_packages(),
    include_package_data=True,    
)

